.rounded, 
.rounded-bottom {
    -moz-border-radius-bottomleft : 8px;
    -moz-border-radius-bottomright : 8px;
    border-bottom-left-radius : 8px;
    border-bottom-right-radius : 8px;
}

.rounded, 
.rounded-top {
    -moz-border-radius-topleft : 8px;
    -moz-border-radius-topright : 8px;
    border-top-left-radius : 8px;
    border-top-right-radius : 8px;
}
