
# This file (at the moment) has the sole purpose of specifying the texttest version number.
# It's in a separate file to make it easy to find.

version = "4.0.8"

# Note: Decided it's not a good idea to require debug versions here. Even if we have Python 2.4.3 it's
# pretty likely 2.4.2 works OK unless we know otherwise. The interface shouldn't be different at least.

# Which python version do we require?
required_python_version = (3, 6, 0)

# Which pygtk version do we require?
required_pygtk_version = (3, 18, 0)
